﻿;;;-*- Mode: Lisp; Package: "OMAS" -*-
;;;===============================================================================
;;;12/11/23
;;;                       E M A I L (file omas-email.lisp)
;;;                       Copyright, Barthès@UTC, 2010-2012
;;;
;;;===============================================================================

;;; This file contains functions to retrieve emails from an imap mailbox and to 
;;; send back HTML MIME messages. Analysis of the incoming mail is ad hoc and very
;;; crude for the time being. It works only for ACL.

;;; The format for emails are the following:
;;; from: <sender>
;;; to: <server> 
;;; subject: <cue>
;;;   {<text containing command and parameters}
;;;
;;; Examples:
;;; from: barthes@utc.fr
;;; to: hdsri@utc.fr
;;; object: hdsri
;;;   contact de barthès au Japon?
;;;
;;; from: barthes@utc.fr
;;; to: hdsri@utc.fr
;;; object: hdsri
;;;   ajouter contact:
;;;     <new contact data>
;;;
;;; IMPORTANT : the object should not contain accentuated characters, otherwise the
;;; MIME encoding takes over and God knows what format it produces...

;;; Available functions are:
;;;
;;;  - SET-MAILBOX-PARAMETERS (server account password)
;;;
;;;  - GET-NEXT-MAIL (agent &key (pattern "[HDSRI]")
;;;       opens the mailbox and returns 2 values:
;;;          - the content of the mail
;;;          - the return address
;;;       or NIL when it fails
;;;
;;;  - PROCESS-NEXT-TASK (agent task-list args)
;;;       takes a list of tasks and executes the first one that has the properties
;;;       where-to-ask and action. 
;;;       args is the list of words to be sent to the agent that will process the task
;;;       obtained usually by merging instructions and email
;;;       Returns 2 values: 
;;;          - :done if a message was sent 
;;;          - the rest of task-list
;;;       or NIL meaning that we could not find a suitable task
;;;
;;;  - PROCESS-SINGLE-EMAIL (agent email)
;;;       inputs the conent part of the email and tries to find a task to apply. If a
;;;       task is found with properties where-to-ask and message-action, then sends a 
;;;       message to the corresponding agent.
;;;       Returns 3 values: 
;;;           - :done if a message was sent to process the email, 
;;;           - list of tasks left to eventually process
;;;           - email cleaned string
;;;         NIL otherwise meaning either that we could not find a task for the message
;;;         or that we could find tasks but none applies.
;;;
;;;  - SEND-HTML-MESSAGE (agent msg &key (subject "De la part de HDSRI"))
;;;       Sends a MIME text/html message to the destination found in the :message-from
;;;       slot of the agent environment

#|
2010
 0617 Creation
 0706 Changing the arguments of some functions to separate subject from rest of email
 0731 returning to original design
2012
 1123 hardcoding smtp1 IP
|#

(in-package :OMAS)

;;; MICROSOFT mailbox mechanism

(eval-when (:compile-toplevel :load-toplevel :execute)
  (require :imap)
  (require :smtp)  ; also loads MIME
  (require :rfc2822)
  (use-package :net.post-office)
  )

#|
(eval-when (:compile-toplevvel :load-toplevel :execute)
  (export 
   '(set-mailbox-parameters get-next-mail process-next-task process-single-email 
                            send-html-message)))
|#
  
;;;===============================================================================
;;; 
;;;                           service macros and globals
;;;
;;;===============================================================================

;(defparameter *email-server* "smtp1.utc.fr")
(defparameter *email-server* "195.83.155.11") ; short DNS
(defparameter *email-user* "hdsri")
(defparameter *email-pwd* "hdsri2010")

;;; used when we get a string containing codes like =E6 in the middle of regular
;;; characters 

(defparameter *French-code-list* 
  '(("=20" . "
")
    ("=
" . " ")
    ("=E0" . "à")
    ("=E2" . "â")
    ("=E4" . "ä")
    ("=E7" . "ç")
    ("=E8" . "è")
    ("=E9" . "é")
    ("=EA" . "ê")
    ("=EB" . "ë")
    ("=EE" . "î")
    ("=EF" . "î")
    ("=F4" . "ô")
    ("=F6" . "ö")
    ("=F9" . "ù")
    ("=FB" . "û")
    ("=FC" . "ü")
    ("=BD" . "oe")
    ))
	
;;;------------------------------------------------------------ extract-email-text
;;; this function should parse MIME messages
;;; handling the instruction part of the message has been disabled because of the
;;; various encodings encountered for the object part of the message from e-mail
;;; tools. Thus, object must contain the a cue, like HDSRI. The rest is ignored.


(defun extract-email-text (agent mb uid)
  "takes an uid and extract info to process message. Removes message ~
    from mailbox.
Argument:
   agent: :HDSRI
   mb: mailbox
   uid: uid corresponding to a mailbox message
Return:
   2 values: a string containing the content of the message
             a string containing the sender's address
   or NIL."
  (declare (ignore agent))
  (let (envelope from-part body input sender-address content)
    ;; get message content
    (setq content (fetch-letter mb uid :uid t))
    ;; get message envelope
    (setq envelope (fetch-parts mb uid "ENVELOPE" :uid t))
    ;(format t "~%; extract-email-text /envelope: ~S" envelope)
    ;; extract from address part, e.g.
    ;;   (("Jean-Paul Barthès" NIL "jean-paul.barthes" "hds.utc.fr"))
    (setq from-part (car (envelope-from (cadddr (cadar envelope)))))
    ;(format t "~%; extract-email-text / (cdr (cadar envelope)): ~S"  
    ;     (cdr (cadar envelope)))
    ;(format t "~%; extract-email-text /from-part: ~S" from-part)
    
    ;; build return-address string from from part
    (setq sender-address (format nil "~A@~A" (nth 2 from-part)(nth 3 from-part)))
    ;(format t "~%; extract-email-text /to-string: ~S" sender-address)
    (format t "~2%<===== email reçu de ~S" sender-address)
    ;; save it for answering message
    ;(env-set agent sender-address :message-from)
    
    ;;--- First check for a MIME multipart message
    (if (is-content-mime? content)
        ;; use ad hoc parsing
        (setq input (get-mime-text/plain content))
      ;; otherwise
      (progn
        ;; get message body
        (setq body (fetch-parts mb uid "body[text]" :uid t))
        ;; get text (we have a single message)
        (setq input (fetch-field (caar body) "body[text]" body))))
    
    ;; clean text
    (setq input (string-trim '(#\space #\newline #\? #\-) input))
    ;(format t "~%; extract-email-text /input: ~S" input)
    (format t "~%<===== ~S" input)
    ;; if input is empty set it to nil
    (if (equal input "")(setq input nil))
    
    ;; remove message from mailbox
    (delete-letter mb uid :uid t)
    ;; return input and sender address
    (values input sender-address)))

;;;------------------------------------------------------------- get-boundary-code

(defun get-boundary-code (email)
  "tries to get boudary code for multipart MIME message.
Agument:
   email: full incoming message (string)
Return: 
   string code or nil"
  (let ((pos (search "boundary=\"" email :test #'string-equal))
        start end)
    (when pos 
      (setq start (+ pos 10))
      (setq end (search "\"" email :start2 start))
      (unless end (return-from get-boundary-code))
      (return-from get-boundary-code (values (subseq email start end) end)))
    ;; try without quotes around boundary code
    (setq pos (search "boundary=" email :test #'string-equal))
    (when pos 
      (setq start (+ pos 9))
      (setq end (search (format nil "~%") email :start2 start))
      (unless end (return-from get-boundary-code))
      (return-from get-boundary-code (values (subseq email start end) end)))
    ;; return nil
    nil))
    
;;;----------------------------------------------------------- get-mime-text/plain

(defun get-mime-text/plain (email)
  "extracts the plain text message from the email text. It is located between the ~
    tag Content-Type: text/plain; and either a boudary mark or the end of the message.
Arguments:
   email: string representing the full email text
   boundary: the boundary mark if any
   start: a position in the email after which we can search (default 0)
Return:
   a string with the content of the message or nil if it fails."
  (let (boundary start pos end)
    ;; first get the boundary mark and location following it
	(multiple-value-setq (boundary start) (get-boundary-code email))
	;; if no boundary mark set start to 0
	(unless boundary (setq start 0))
    ;; get the text/plain mark
    (setq pos (search "text/plain" email :test #'string-equal :start2 start))
    ;; if nil we quit
    (unless pos (return-from get-mime-text/plain))
    
    ;; locate next Content-Transfer-Encoding
    (setq pos (search "Content-Transfer-Encoding" email :test #'string-equal 
                      :start2 pos))
    ;; if nil we quit
    (unless pos (return-from get-mime-text/plain))
    
    ;; if OK get to the end of the line
    (setq pos (or (position #\return email :start pos)
                  (position #\linefeed email :start pos)))
    ;; now text is either between here and boudary or the end of email
    (if boundary (setq end (search boundary email :start2 pos)))
	(format t "~% get-mime-text/plain /end of text: ~S" end)
    
    ;; if end is nil we extract everything to the end of the message
    (string-retrieve-accents (subseq email pos end))))     
        
;;;---------------------------------------------------------------- get-next-email

(defun get-next-email (agent &key (pattern "[HDSRI]"))
  "opens the mailbox and returns the next email available from the mailbox ~
   with object containing pattern. If none, empties the mailbox and return ~
   nil. 
Argument:
   agent: PA_HDSRI
   pattern (key): string to be found in the object part of email, default \"[HDSRI]\"
Return:
   3 values: a string containing the instruction part of the subject
             a string corresponding to the content of the message 
             a string containing the sender's address
   or NIL."
  (declare (special *email-server* *email-user* *email-pwd*))
  (let (msg-list)
    (with-imap-connection 
        (mb *email-server* :user *email-user* :password *email-pwd*)
      ;; select input mailbox
      (select-mailbox mb "inbox")
      ;; get query messages, returning list of messages UIDs
      (setq msg-list (search-mailbox mb `(:subject ,pattern) :uid t))
      ;; if no such messages, eventually clear mailbox
      (unless msg-list 
        ;; here we clear the mailbox to get rid of spam
        (setq msg-list (search-mailbox mb :all :uid t))
        (when msg-list
          (format t "~%+++++ deleting ~S irrelevant ~A email(s)." 
            (length msg-list) pattern)
          (dolist (item msg-list)
            (delete-letter mb item :uid t)))		
        (return-from get-next-email nil))
      
      ;; get first message and return it (message is deleted from mailbox)
      ;; returns 2 values: message-content and sender address
      (extract-email-text agent mb (car msg-list)))))
      
;;;-------------------------------------------------------------- is-content-MIME?

(defun is-content-MIME? (email)
   "checks crudely if content is MIME by trying to locate MIME-Version in string.
Argument: 
   email (string)
Return:
   NIL if not MIME."
  (and (search "MIME-Version" email :test #'string-equal)
       (get-boundary-code email)))

;;;------------------------------------------------------------- process-next-task
;;; moved to API

;;;(defun  process-next-task (agent task-list args)
;;;  "takes a list of tasks and executes the first one that has the properties ~
;;;    where-to-ask and action. Remaining task list is returned as second value.
;;;Arguments:	
;;;   agent: current agent
;;;   task-list: a list of tasks
;;;   args: arguments for the message to be sent
;;;Return:
;;;   if a message was sent returns 2 values :done and rest of task-list, 
;;;   nil otherwise"
;;;  (when task-list
;;;    (let ((selected-task (car task-list))
;;;          to-arg action-arg)
;;;      (loop
;;;        ;; for the next line to execute one must be inside a task process...
;;;        (when (and (setq to-arg (car (send selected-task '=get "where to ask")))
;;;                   (setq action-arg (car (send selected-task '=get "message action"))))
;;;          (format t "~%; process-next-task /activating subtask: ~S" selected-task)
;;;          ;; set pattern to (:pattern . :html) to ask for an html string in return
;;;          ;(setq args (cons '(:pattern . :html) (remove :pattern args :key #'car)))
;;;          (send-subtask agent :to to-arg :action action-arg :args (list args))
;;;          ;; save rest of task-list
;;;          ;(env-set agent (cdr task-list) :next-tasks)
;;;          (return-from process-next-task (values :done (cdr task-list))))
;;;        
;;;        (format t "~%; process-next-task / task ~S failed." selected-task)
;;;        
;;;        ;; if task-list empty return
;;;        (unless task-list 
;;;          (format t "~%; process-next-task /no more task we quit.")
;;;          (return-from process-next-task))
;;;        
;;;        ;; otherwise pop pair list
;;;        (setq selected-task (pop task-list))
;;;        ;; go to the start of the loop
;;;        )
;;;      )))

;;;---------------------------------------------------------- process-single-email
;;; for message in which we want to add something, we should
;;;  - first get initial text up to first ":"
;;;  - do the task analysis on that
;;;  - pass the rest of arguments as a single string rather than a list of words
;;; meaning that: all messages that contain ":" are assumed to have a first part
;;; specifying the task to do and a second part containing parameters that will
;;; be parsed by the task.
;;; Another solution swould be to pass the message content as a single string for
;;; all automatic messages

(defUn process-single-email (agent email)
  "takes a string and tries to make sense out of it, i.e. find the task ~
  and ship the request to the right agent.
Arguments:
   agent: HDSRI
   email: the string from the message content
Returns:
   3 values: :done if a message was sent to process the email, 
             list of tasks left to eventually process
             email cleaned string
   NIL otherwise meaning either that we could not find a task for the message
   or that we could find tasks but none applies."
  (let (input task-list patterns level weights word-weight-list
              result first-task pair-list message-sent? raw-input
              pos instructions)
    ;;=== process first instructions
    ;; first remove some quote-printable codes
    (setq raw-input (string-retrieve-accents email))
    (format t "~%; process-single-email /translated email: ~S" raw-input)
    
    ;; we check to see if the text contains a first part corresponding to
    ;; a command, e.g. "add contact :" which would be marked by a ":" sign
    (setq pos (position #\: raw-input))
    ;; if so, extract first part, otherwise use all
    (setq instructions 
          (if pos (subseq raw-input 0 pos) raw-input))
    ;; now we will use instructions to locate a task
    
    ;; preprocess email string -> list of words
    ;; must remove final question mark or period
    (setq input (moss::%get-words-from-text (string-trim '(#\. #\?) instructions) 
                                            moss::*delimiters*))
    ;; we assume that resulting text contains a request or command
    ;; try to find a task from the input text
    (setq task-list (moss::find-objects 
                     '("task" ("index-pattern" ("task-index" ("index" :is :?))))
                     input :all-objects t))
    ;; if no task quit
    (unless task-list
      (return-from process-single-email nil))
    
    ;;=== select task to activate
    ;; first compute a list of patterns (combinations of words) from the input
    (setq patterns (mapcar #'car (moss::generate-access-patterns input)))
    (moss::vformat "process-single-email /input: ~S~&  generated patterns:~&~S" 
                   input patterns)
    ;; then, for each task
    (dolist (task task-list)
      (setq level 0)
      ;; get the weight list
      (setq weights (moss::%get-INDEX-WEIGHTS task))
      (moss::vformat "process-single-email /task: ~S~&  weights: ~S" task weights)
      ;; check the patterns according to the weight list
      (setq word-weight-list (moss::%get-relevant-weights weights patterns))
      (moss::vformat "process-single-email /word-weight-list:~&  ~S" word-weight-list)
      ;; combine the weights
      (dolist (item word-weight-list)
        (setq level (+ level (cadr item) (- (* level (cadr item))))))
      (moss::vformat "process-single-email /level: ~S" level)
      ;; push the task and weight onto the result list
      (push (list task level) result)
      )
    (moss::vformat "process-single-email /result:~&~S" result)
    ;; if no task applies return nil
    (unless result
      (return-from process-single-email))
    
    ;; otherwise, order the list
    (setq pair-list (sort result #'> :key #'cadr))
    (moss::vformat "process-single-email /pair-list:~&  ~S" pair-list)
    
    ;; keep the first task even if its score is low
    (setq first-task (caar pair-list))
    ;; remove the task that have a weight less than task-threshold (default 0.4)
    (setq pair-list
          (remove nil 
                  (mapcar 
                      #'(lambda (xx) 
                          (if (>= (cadr xx) (omas::task-threshold omas::*omas*)) xx))
                    pair-list)))
    ;; compute list of tasks
    (setq task-list (mapcar #'car pair-list))
    
    ;; activate first eligible task, answer will be processed by dynamic 
    ;; part of skill
    (multiple-value-setq (message-sent? task-list)
      (process-next-task agent (or task-list (list first-task))
                         ;`((:data . ,input)(:language . :fr))))
                         `((:raw-data . ,raw-input)(:data . ,input)(:language . :fr)))) ; jpb 1008
    ;; returns :done if message was sent, nil otherwise, meaning that no task applies
    (values message-sent? task-list raw-input)))

#|
? (omas::process-single-email hdsri::PA_hdsri "contacts avec la chine?")
? (process-single-mail hdsri::PA_hdsri "stats projets?" "")
|#
;;;------------------------------------------------------------- send-html-message

(defun send-html-message (agent text to-address &key (subject "No subject"))
  "sends a message to the caller using an html string. To-address is
    kept in the :message-from slot of the agent. Synchronous send.
Argument:
   agent: PA_HDSRI
   text: html string 
   to-address: a string
   subject (key): (default : De la part de HDSRI)
Return:
   :done or nil in case of error somewhere"
  (declare (special *email-server*))
  (let ((msg (make-mime-part :text text :content-type "text/html")))
    ;; send it back
    (send-letter *email-server* 
                 (format nil "~A@utc.fr" (key agent)) to-address msg 
                 :subject subject)
    (format t "~%; send-html-message /answer sent to: ~S" to-address)
    :done))

#|
(send-letter "195.83.155.11" "hdsri@utc.fr" "barthes@utc.fr" "Coucou...5")
(send-letter "smtp1.utc.fr" "hdsri@utc.fr" "barthes@utc.fr" "Coucou...5")
(send-letter "195.83.155.11" "MISSION@utc.fr" "barthes@utc.fr" "Coucou...6")
(send-html-message mission::sa_mission "Youpee..." "barthes@utc.fr")
|#
;;;-------------------------------------------------------- set-mailbox-parameters

(defun set-mailbox-parameters (server user password)
  "self explanatory"
   (declare (special *email-server* *email-user* *email-pwd*))
   (setq *email-server* server
         *email-user* user
         *email-pwd* password))
	  
;;;------------------------------------------------------- string-retrieve-accents

(defun string-retrieve-accents (text)
  "kludge to recover accents from email strings."
  (let ((code-list *french-code-list*) (pos 0) code)
    (loop
      (unless code-list (return-from string-retrieve-accents text))
      (setq code (caar code-list))
      (loop
        (setq pos (search code text))
        (unless pos (return))
        ;; replace code with spécial letter
        (setq text (concatenate 'string (subseq text 0 pos) (cdar code-list)
                     (subseq text (+ pos 3))))
        )
      (pop code-list))))

#|
(string-retrieve-accents "barth=E8s")
"barthès"
(string-retrieve-accents "fran=E7ois")
"françois"
|#	 

(format t "~&;*** OMAS v~A - email loaded ***" *omas-version-number*)
:EOF